from isitfit.cost.ec2_analyze import pipeline_factory as ec2_cost_analyze
from isitfit.cost.ec2_optimize import pipeline_factory as ec2_cost_optimize

from isitfit.cost.redshift_analyze import pipeline_factory as redshift_cost_analyze
from isitfit.cost.redshift_optimize import pipeline_factory as redshift_cost_optimize

from isitfit.cost.account_cost_analyze  import pipeline_factory as account_cost_analyze
from isitfit.cost.account_cost_optimize import pipeline_factory as account_cost_optimize

